from os import makedirs
from pathlib import Path
from subprocess import check_call
from tempfile import TemporaryDirectory
from typing import Generator

import pytest
import requests
import yaml

from tests.integration_tests import ADMIN_TOKEN


@pytest.yield_fixture(scope="session")
def local_grand_challenge() -> Generator[str, None, None]:
    local_api_url = "https://gc.localhost/api/v1/"

    try:
        r = requests.get(
            local_api_url,
            verify=False,
            headers={"Authorization": f"TOKEN {ADMIN_TOKEN}"},
        )
        r.raise_for_status()
        local_gc_running = True
    except requests.exceptions.ConnectionError:
        local_gc_running = False

    if local_gc_running:
        yield local_api_url
    else:
        # Start our own version of grand challenge
        with TemporaryDirectory() as tmp_path:

            for f in [
                "docker-compose.yml",
                "dockerfiles/db/postgres.test.conf",
            ]:
                get_grand_challenge_file(Path(f), Path(tmp_path))

            try:
                check_call(["docker-compose", "pull"], cwd=tmp_path)

                for command in [
                    "migrate",
                    "check_permissions",
                    "init_gc_demo",
                ]:
                    check_call(
                        [
                            "docker-compose",
                            "run",
                            "--rm",
                            "web",
                            "python",
                            "manage.py",
                            command,
                        ],
                        cwd=tmp_path,
                    )
                check_call(["docker-compose", "up", "-d"], cwd=tmp_path)
                check_call(
                    ["docker-compose-wait", "-w", "-t", "2m"], cwd=tmp_path
                )

                yield local_api_url

            finally:
                check_call(["docker-compose", "down"], cwd=tmp_path)


def get_grand_challenge_file(repo_path: Path, output_directory: Path) -> None:
    r = requests.get(
        (
            f"https://raw.githubusercontent.com/comic/grand-challenge.org/"
            f"master/{repo_path}"
        ),
        allow_redirects=True,
    )

    if str(repo_path) == "docker-compose.yml":
        content = rewrite_docker_compose(r.content)
    else:
        content = r.content

    output_file = output_directory / repo_path
    makedirs(str(output_file.parent), exist_ok=True)

    with open(str(output_file), "wb") as f:
        f.write(content)


def rewrite_docker_compose(content: bytes) -> bytes:
    spec = yaml.safe_load(content)

    for s in spec["services"]:
        # Remove the non-postgres volume mounts, these are not needed for testing
        if s != "postgres" and "volumes" in spec["services"][s]:
            del spec["services"][s]["volumes"]

        # Replace test with production containers
        if spec["services"][s]["image"] == "grandchallenge/web-test:latest":
            spec["services"][s]["image"] = "grandchallenge/web:latest"

    # Use the production web server as the test one is not included
    spec["services"]["web"][
        "command"
    ] = "gunicorn -b 0.0.0.0 -k uvicorn.workers.UvicornWorker config.asgi:application"

    return yaml.safe_dump(spec).encode("utf-8")



import itertools
import os
import uuid
import warnings
from collections import UserDict
from io import BytesIO
from json import load
from pathlib import Path
from random import randint, random
from time import sleep, time
from typing import Dict, List, Type
from urllib.parse import urljoin

import jsonschema
from requests import ConnectionError, Session


def is_uuid(s):
    try:
        uuid.UUID(s)
    except ValueError:
        return False
    else:
        return True


def accept_tuples_as_arrays(org):
    return org.redefine(
        "array",
        lambda checker, instance: isinstance(instance, tuple)
        or org.is_type(instance, "array"),
    )


Draft7ValidatorWithTupleSupport = jsonschema.validators.extend(
    jsonschema.Draft7Validator,
    type_checker=accept_tuples_as_arrays(
        jsonschema.Draft7Validator.TYPE_CHECKER
    ),
)


class LazyDict(UserDict):
    def __init__(self, load_function):
        self.__load_function = load_function
        self.__data_loaded = False
        self.__data = None

    def load(self):
        if not self.__data_loaded:
            self.__data = self.__load_function()
            self.__data_loaded = True

    @property
    def data(self):
        self.load()
        return self.__data


def load_input_data(input_file):
    with open(str(Path(__file__).parent / input_file), "rb") as f:
        return f.read()


def generate_new_upload_id(content):
    return "{}_{}_{}".format(hash(content), time(), random())


def import_json_schema(filename):
    """
    Loads a json schema from the module's subdirectory "schemas".

    This is not *really* an import but the naming indicates that an ImportError
    is raised in case the json schema cannot be loaded. This should also only
    be called while the module is loaded, not at a later stage, because import
    errors should be raised straight away.

    Parameters
    ----------
    filename: str
        The jsonschema file to be loaded. The filename is relative to the
        "schemas" directory.

    Returns
    -------
    Draft7Validator
        The jsonschema validation object

    Raises
    ------
    ImportError
        Raised if the json schema cannot be loaded.
    """
    filename = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "schemas", filename
    )

    try:
        with open(filename) as f:
            jsn = load(f)
        return Draft7ValidatorWithTupleSupport(
            jsn, format_checker=jsonschema.draft7_format_checker
        )
    except ValueError as e:
        # I want missing/failing json imports to be an import error because that
        # is what they should indicate: a "broken" library
        raise ImportError(
            "Json schema '{file}' cannot be loaded: {error}".format(
                file=filename, error=e
            )
        )


class APIBase:
    _client = None  # type: Client
    base_path = ""
    sub_apis: Dict[str, Type["APIBase"]] = {}

    json_schema = None

    def __init__(self, client):
        if isinstance(self, ModifiableMixin):
            ModifiableMixin.__init__(self)

        self._client = client

        for k, api in list(self.sub_apis.items()):
            setattr(self, k, api(self._client))

    def _verify_against_schema(self, value):
        if self.json_schema is not None:
            self.json_schema.validate(value)

    def list(self, params=None):
        result = self._client(method="GET", path=self.base_path, params=params)
        for i in result:
            self._verify_against_schema(i)
        return result

    def page(self, offset=0, limit=100, params=None):
        if params is None:
            params = {}
        params["offset"] = offset
        params["limit"] = limit
        result = self._client(
            method="GET", path=self.base_path, params=params,
        )["results"]
        for i in result:
            self._verify_against_schema(i)
        return result

    def iterate_all(self, params=None):
        req_count = 100
        offset = 0
        while True:
            current_list = self.page(
                offset=offset, limit=req_count, params=params
            )
            if len(current_list) == 0:
                break
            yield from current_list
            offset += req_count

    def detail(self, pk):
        result = self._client(
            method="GET", path=urljoin(self.base_path, pk + "/")
        )
        self._verify_against_schema(result)
        return result


class ModifiableMixin:
    _client = None  # type: Client

    modify_json_schema = None  # type: jsonschema.Draft7Validator

    def __init__(self):
        pass

    def _process_post_arguments(self, post_args):
        if self.modify_json_schema is not None:
            self.modify_json_schema.validate(post_args)

    def _validate_data(self, data):
        if data is None:
            data = {}
        self._process_post_arguments(data)
        return data

    def _execute_request(self, method, data, pk):
        url = (
            self.base_path
            if not pk
            else urljoin(self.base_path, str(pk) + "/")
        )
        return self._client(method=method, path=url, json=data)

    def perform_request(self, method, data=None, pk=False, validate=True):
        if validate:
            data = self._validate_data(data)
        return self._execute_request(method, data, pk)

    def send(self, **kwargs):
        # Created for backwards compatibility
        self.create(**kwargs)

    def create(self, **kwargs):
        return self.perform_request(
            "POST", data=kwargs, validate=kwargs.get("validate", True)
        )

    def update(self, pk, **kwargs):
        return self.perform_request("PUT", pk=pk, data=kwargs)

    def partial_update(self, pk, **kwargs):
        return self.perform_request("PATCH", pk=pk, data=kwargs)

    def delete(self, pk):
        return self.perform_request("DELETE", pk=pk, validate=False)


class ImagesAPI(APIBase):
    base_path = "cases/images/"


class UploadSessionFilesAPI(APIBase, ModifiableMixin):
    base_path = "cases/upload-sessions/files/"


class UploadSessionsAPI(APIBase, ModifiableMixin):
    base_path = "cases/upload-sessions/"

    def process_images(self, pk, json=None):
        url = urljoin(self.base_path, str(pk) + "/process_images/")
        return self._client(method="PATCH", path=url, json=json)


class WorkstationSessionsAPI(APIBase):
    base_path = "workstations/sessions/"


class ReaderStudyQuestionsAPI(APIBase):
    base_path = "reader-studies/questions/"


class ReaderStudyMineAnswersAPI(APIBase, ModifiableMixin):
    base_path = "reader-studies/answers/mine/"
    json_schema = import_json_schema("answer.json")


class ReaderStudyAnswersAPI(APIBase, ModifiableMixin):
    base_path = "reader-studies/answers/"
    json_schema = import_json_schema("answer.json")
    modify_json_schema = import_json_schema("post-answer.json")

    sub_apis = {"mine": ReaderStudyMineAnswersAPI}

    mine = None  # type: ReaderStudyMineAnswersAPI

    def _process_post_arguments(self, post_args):
        if is_uuid(post_args["question"]):
            post_args["question"] = urljoin(
                urljoin(
                    self._client.base_url, ReaderStudyQuestionsAPI.base_path
                ),
                post_args["question"] + "/",
            )

        ModifiableMixin._process_post_arguments(self, post_args)


class ReaderStudiesAPI(APIBase):
    base_path = "reader-studies/"
    json_schema = import_json_schema("reader-study.json")

    sub_apis = {
        "answers": ReaderStudyAnswersAPI,
        "questions": ReaderStudyQuestionsAPI,
    }

    answers = None  # type: ReaderStudyAnswersAPI
    questions = None  # type: ReaderStudyQuestionsAPI

    def detail(self, pk):
        result = LazyDict(lambda: super().detail(pk))
        result.ground_truth = GroundTruthAPI(self._client, pk)
        return result

    def ground_truth(self, pk, case_pk):
        warnings.warn(
            DeprecationWarning(
                "Please use reader_studies.detail(pk).ground_truth.detail(case_pk) "
                "instead of reader_studies.ground_truth(self, pk, case_pk)"
            )
        )
        return self.detail(pk).ground_truth.detail(case_pk)


class GroundTruthAPI(APIBase):
    def __init__(self, client, pk):
        self.base_path = urljoin(
            ReaderStudiesAPI.base_path, f"{pk}/ground-truth/"
        )

        super().__init__(client)


class AlgorithmsAPI(APIBase):
    base_path = "algorithms/"


class AlgorithmResultsAPI(APIBase):
    base_path = "algorithms/results/"


class AlgorithmJobsAPI(APIBase):
    base_path = "algorithms/jobs/"

    def by_input_image(self, pk):
        return self.iterate_all(params={"image": pk})


class RetinaLandmarkAnnotationSetsAPI(APIBase, ModifiableMixin):
    base_path = "retina/landmark-annotation/"
    json_schema = import_json_schema("landmark-annotation.json")
    modify_json_schema = import_json_schema("post-landmark-annotation.json")

    def for_image(self, pk):
        result = self._client(
            method="GET", path=self.base_path, params={"image_id": pk}
        )
        for i in result:
            self._verify_against_schema(i)
        return result


class RetinaPolygonAnnotationSetsAPI(APIBase, ModifiableMixin):
    base_path = "retina/polygon-annotation-set/"
    json_schema = import_json_schema("polygon-annotation.json")
    modify_json_schema = import_json_schema("post-polygon-annotation.json")


class RetinaSinglePolygonAnnotationsAPI(APIBase, ModifiableMixin):
    base_path = "retina/single-polygon-annotation/"
    json_schema = import_json_schema("single-polygon-annotation.json")
    modify_json_schema = import_json_schema(
        "post-single-polygon-annotation.json"
    )


class ChunkedUploadsAPI(APIBase):
    base_path = "chunked-uploads/"

    def _upload_file_with_exponential_backoff(self, file_info):
        """
        Uploads a chunk with an exponential backoff retry strategy. The
        maximum number of attempts is 3.

        Parameters
        ----------
        file_info: dict
        Contains information about the chunk, start_byte, end_byte and upload_id.

        Raises
        ------
        ConnectionError
            Raised if the chunk cannot be uploaded within 3 attempts.
        """
        num_retries = 0
        e = Exception
        while num_retries < 3:
            try:
                result = self._client(
                    method="POST",
                    path=self.base_path,
                    files={file_info["filename"]: BytesIO(file_info["chunk"])},
                    data={
                        "filename": file_info["filename"],
                        "X-Upload-ID": file_info["upload_id"],
                    },
                    extra_headers={
                        "Content-Range": "bytes {}-{}/{}".format(
                            file_info["start_byte"],
                            file_info["end_byte"] - 1,
                            len(file_info["content"]),
                        )
                    },
                )
                break
            except ConnectionError as _e:
                num_retries += 1
                e = _e
                sleep((2 ** num_retries) + (randint(0, 1000) / 1000))
        else:
            raise e

        return result

    def upload_file(self, filename):
        """
        Uploads a file in chunks using rest api.

        Parameters
        ----------
        filename: str
            The name of the file to be uploaded.

        Raises
        ------
        ConnectionError
            Raised if a chunk cannot be uploaded.
        """

        content = load_input_data(filename)
        upload_id = generate_new_upload_id(content)
        start_byte = 0
        content_io = BytesIO(content)
        max_chunk_length = 2 ** 23
        results = []

        while True:
            chunk = content_io.read(max_chunk_length)

            if not chunk:
                break

            end_byte = start_byte + len(chunk)

            try:
                result = self._upload_file_with_exponential_backoff(
                    {
                        "start_byte": start_byte,
                        "end_byte": end_byte,
                        "chunk": chunk,
                        "content": content,
                        "upload_id": upload_id,
                        "filename": str(filename),
                    }
                )
            except ConnectionError as e:
                raise e

            results.append(result)

            start_byte += len(chunk)

        return list(itertools.chain(*results))


class WorkstationConfigsAPI(APIBase):
    base_path = "workstations/configs/"


class Client(Session):
    def __init__(
        self,
        token: str = "",
        base_url: str = "https://grand-challenge.org/api/v1/",
        verify: bool = True,
    ):
        super().__init__()

        self.headers.update({"Accept": "application/json"})

        if token:
            self.headers.update({"Authorization": f"TOKEN {token}"})
        else:
            token = str(os.getenv("GRAND_CHALLENGE_AUTHORIZATION", ""))
            if token:
                # Already contains TOKEN prefix
                self.headers.update({"Authorization": token})
            else:
                raise RuntimeError("Token must be set")

        self._base_url = base_url
        if not self._base_url.startswith("https://"):
            raise RuntimeError("Base URL must be https")

        # Should we verify the servers SSL certificates?
        self._verify = verify

        self.images = ImagesAPI(client=self)
        self.reader_studies = ReaderStudiesAPI(client=self)
        self.sessions = WorkstationSessionsAPI(client=self)
        self.chunked_uploads = ChunkedUploadsAPI(client=self)
        self.algorithms = AlgorithmsAPI(client=self)
        self.algorithm_results = AlgorithmResultsAPI(client=self)
        self.algorithm_jobs = AlgorithmJobsAPI(client=self)
        self.workstation_configs = WorkstationConfigsAPI(client=self)
        self.retina_landmark_annotations = RetinaLandmarkAnnotationSetsAPI(
            client=self
        )
        self.retina_polygon_annotation_sets = RetinaPolygonAnnotationSetsAPI(
            client=self
        )
        self.retina_single_polygon_annotations = RetinaSinglePolygonAnnotationsAPI(
            client=self
        )
        self.raw_image_upload_session_files = UploadSessionFilesAPI(
            client=self
        )
        self.raw_image_upload_sessions = UploadSessionsAPI(client=self)

    @property
    def base_url(self):
        return self._base_url

    def _validate_url(self, url):
        if not url.startswith(self._base_url):
            raise RuntimeError(f"{url} does not start with {self._base_url}")

    def __call__(
        self,
        method="GET",
        url="",
        path="",
        params=None,
        json=None,
        extra_headers=None,
        files=None,
        data=None,
    ):
        if not url:
            url = urljoin(self._base_url, path)
        if extra_headers is None:
            extra_headers = {}
        if json is not None:
            extra_headers["Content-Type"] = "application/json"

        self._validate_url(url)
        response = self.request(
            method=method,
            url=url,
            files={} if files is None else files,
            data={} if data is None else data,
            headers={**self.headers, **extra_headers},
            verify=self._verify,
            params={} if params is None else params,
            json=json,
        )
        response.raise_for_status()
        if response.headers.get("Content-Type") == "application/json":
            return response.json()
        else:
            return response

    def run_external_algorithm(
        self, algorithm_name: str, files_to_upload: List[str]
    ) -> Dict:
        """
        This function uploads an input image to grand challenge and runs an
        already uploaded algorithm which the user has access to. If the upload
        is finished correctly, grand challenge will automatically submit a job
        which you can access via AlgorithmJobsAPI.

        After the job is finished successfully the result will be available via
        AlgorithmResultsAPI.

        Parameters
        ----------
        algorithm_name:
            Title of the algorithm which has already been uploaded.
        files_to_upload:
            List of files to upload.

        Returns
        -------
        upload_session:
        This can be used to construct a query like
        /api/v1/cases/images/?origin=upload_session["pk"]
        to find out which Image did this RawImageUploadSession give rise to.
        This can be further used to identify the submitted job.
        """
        return self.upload_cases(
            algorithm=algorithm_name, files_to_upload=files_to_upload
        )

    def upload_cases(
        self,
        *,
        files_to_upload: List[str],
        reader_study: str = None,
        archive: str = None,
        algorithm: str = None,
    ):
        upload_session_data = {}

        if reader_study is not None:
            upload_session_data["reader_study"] = reader_study

        if archive is not None:
            upload_session_data["archive"] = archive

        if algorithm is not None:
            upload_session_data[
                "algorithm_image"
            ] = self._get_latest_algorithm_image(algorithm_name=algorithm)

        raw_image_upload_session = self.raw_image_upload_sessions.create(
            **upload_session_data
        )

        uploaded_files = {}
        for file in files_to_upload:
            uploaded_chunks = self.chunked_uploads.upload_file(file)
            uploaded_files.update(
                {c["uuid"]: c["filename"] for c in uploaded_chunks}
            )

        for file_id, filename in uploaded_files.items():
            self.raw_image_upload_session_files.create(
                upload_session=raw_image_upload_session["api_url"],
                staged_file_id=file_id,
                filename=filename,
            )

        self.raw_image_upload_sessions.process_images(
            pk=raw_image_upload_session["pk"]
        )

        return raw_image_upload_session

    def _get_latest_algorithm_image(self, algorithm_name: str) -> str:
        """Get the latest algorithm image for the given algorithm name. """
        algorithms = [
            a
            for a in self.algorithms.list()["results"]
            if a["title"] == algorithm_name
        ]

        if len(algorithms) != 1:
            raise ValueError(
                "{} is not found in available list of algorithms".format(
                    algorithm_name
                )
            )

        if not algorithms[0]["latest_ready_image"]:
            raise ValueError(f"{algorithm_name} is not ready to be used")

        algorithm_image = algorithms[0]["latest_ready_image"]

        return algorithm_image



import os

import pytest
from click.testing import CliRunner
from jsonschema import ValidationError

from gcapi import Client, cli
from gcapi.gcapi import LazyDict


def test_lazy_dict():
    ld = LazyDict(lambda: {1: 2})
    assert len(ld) == 1

    ld = LazyDict(lambda: {1: 2})
    assert ld[1] == 2


def test_no_auth_exception():
    with pytest.raises(RuntimeError):
        Client()


def test_headers():
    token = "foo"
    c = Client(token=token)
    assert c.headers["Authorization"] == f"TOKEN {token}"
    assert c.headers["Accept"] == "application/json"


def test_token_via_env_var():
    token = "TOKEN 1b9436200001f2eaf57cd77db075cbb60a49a00a"
    os.environ["GRAND_CHALLENGE_AUTHORIZATION"] = token
    c = Client()
    assert c.headers["Authorization"] == token


def test_http_base_url():
    with pytest.raises(RuntimeError):
        Client(token="foo", base_url="http://example.com")


def test_custom_base_url():
    c = Client(token="foo")
    assert c._base_url.startswith("https://grand-challenge.org")

    c = Client(token="foo", base_url="https://example.com")
    assert c._base_url.startswith("https://example.com")


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.main)
    assert result.exit_code == 0
    assert "gcapi.cli.main" in result.output
    help_result = runner.invoke(cli.main, ["--help"])
    assert help_result.exit_code == 0
    assert "--help  Show this message and exit." in help_result.output


@pytest.mark.parametrize(
    "datetime_string,valid",
    (
        ("teststring", False),
        (1, False),
        ({}, False),
        ("2019-13-11T13:55:00.123456Z", False),
        ("2019-12-11T25:55:00.123456Z", False),
        ("2019-12-11T13:60:00.123456Z", False),
        ("2019-12-11T13:55:00.123456Z", True),
        ("2019-12-11T13:55:00Z", True),
    ),
)
def test_datetime_string_format_validation(datetime_string, valid):
    landmark_annotation = {
        "id": "4d5721f8-485d-4a17-8507-e06a8f897dd3",
        "grader": 7,
        "created": datetime_string,
        "singlelandmarkannotation_set": [
            {
                "id": "69ea96c2-9a96-4080-9b2c-0b9d0417dda1",
                "image": "70ec13fd-7fcf-4c84-bcd0-5fa3ac34a6b0",
                "landmarks": [
                    [249.029700179422, 194.950491966439],
                    [308.910901038675, 210.158411188267],
                    [270.891081357472, 353.683160558702],
                ],
            },
            {
                "id": "8055d041-d78d-4fa6-94af-1e57e80e11d8",
                "image": "46ba46f4-7fcd-418d-a43f-f668b286daeb",
                "landmarks": [
                    [759.567661360211, 495.505389057573],
                    [911.79182925945, 646.176269350096],
                    [672.582428997143, 726.948261175343],
                ],
            },
        ],
    }
    c = Client(verify=False, token="foo")
    if valid:
        assert (
            c.retina_landmark_annotations._verify_against_schema(
                landmark_annotation
            )
            is None
        )
    else:
        with pytest.raises(ValidationError):
            c.retina_landmark_annotations._verify_against_schema(
                landmark_annotation
            )



import os
from typing import Dict

from setuptools import find_packages, setup

NAME = "gcapi"

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "Click>=6.0",
    "Requests",
    "jsonschema[format_nongpl]>=3.0",
]

test_requirements = [
    "pytest",
    "pyyaml",
    "docker-compose-wait",
    "pytest-cov",
]

about: Dict[str, str] = {}
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, NAME, "__version__.py")) as f:
    exec(f.read(), about)

setup(
    author="James Meakin",
    author_email="code@jmsmkn.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Python client for the grand-challenge.org API",
    entry_points={"console_scripts": ["gcapi=gcapi.cli:main"]},
    extras_require={"test": test_requirements},
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="gcapi",
    name=NAME,
    packages=find_packages(include=["gcapi"]),
    package_data={"gcapi": ["schemas/*"]},
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/DIAGNijmegen/gcapi",
    version=about["__version__"],
    zip_safe=False,
)



import os
from pathlib import Path
from time import sleep

import pytest
from requests import HTTPError

from gcapi import Client

RETINA_TOKEN = "f1f98a1733c05b12118785ffd995c250fe4d90da"
ADMIN_TOKEN = "1b9436200001f2eaf57cd77db075cbb60a49a00a"
ALGORITHMUSER_TOKEN = "dc3526c2008609b429514b6361a33f8516541464"


@pytest.mark.parametrize(
    "annotation",
    [
        "retina_landmark_annotations",
        "retina_polygon_annotation_sets",
        "retina_single_polygon_annotations",
    ],
)
def test_list_annotations(local_grand_challenge, annotation):
    c = Client(
        base_url=local_grand_challenge, verify=False, token=RETINA_TOKEN
    )
    response = getattr(c, annotation).list()
    assert len(response) == 0


def test_create_landmark_annotation(local_grand_challenge):
    c = Client(
        base_url=local_grand_challenge, verify=False, token=RETINA_TOKEN
    )
    nil_uuid = "00000000-0000-4000-9000-000000000000"
    create_data = {
        "grader": 0,
        "singlelandmarkannotation_set": [
            {"image": nil_uuid, "landmarks": [[0, 0], [1, 1], [2, 2]]},
            {"image": nil_uuid, "landmarks": [[0, 0], [1, 1], [2, 2]]},
        ],
    }
    with pytest.raises(HTTPError) as e:
        c.retina_landmark_annotations.create(**create_data)
    response = e.value.response
    assert response.status_code == 400
    response = response.json()
    assert response["grader"][0] == 'Invalid pk "0" - object does not exist.'
    for sla_error in response["singlelandmarkannotation_set"]:
        assert (
            sla_error["image"][0]
            == f'Invalid pk "{nil_uuid}" - object does not exist.'
        )


def test_create_polygon_annotation_set(local_grand_challenge):
    c = Client(
        base_url=local_grand_challenge, verify=False, token=RETINA_TOKEN
    )
    nil_uuid = "00000000-0000-4000-9000-000000000000"
    create_data = {
        "grader": 0,
        "image": nil_uuid,
        "singlepolygonannotation_set": [
            {"z": 0, "value": [[0, 0], [1, 1], [2, 2]]},
            {"z": 1, "value": [[0, 0], [1, 1], [2, 2]]},
        ],
    }
    with pytest.raises(HTTPError) as e:
        c.retina_polygon_annotation_sets.create(**create_data)
    response = e.value.response
    assert response.status_code == 400
    response = response.json()
    assert response["grader"][0] == 'Invalid pk "0" - object does not exist.'
    assert (
        response["image"][0]
        == f'Invalid pk "{nil_uuid}" - object does not exist.'
    )
    assert response["name"][0] == "This field is required."


def test_create_single_polygon_annotations(local_grand_challenge):
    c = Client(
        base_url=local_grand_challenge, verify=False, token=RETINA_TOKEN
    )
    create_data = {
        "z": 0,
        "value": [[0, 0], [1, 1], [2, 2]],
        "annotation_set": 0,
    }

    with pytest.raises(HTTPError) as e:
        c.retina_single_polygon_annotations.create(**create_data)
    response = e.value.response
    assert response.status_code == 400
    response = response.json()
    assert (
        response["annotation_set"][0]
        == 'Invalid pk "0" - object does not exist.'
    )


def test_raw_image_and_upload_session(local_grand_challenge):
    c = Client(base_url=local_grand_challenge, verify=False, token=ADMIN_TOKEN)
    assert c.raw_image_upload_session_files.page() == []
    assert c.raw_image_upload_sessions.page() == []


def test_local_response(local_grand_challenge):
    c = Client(base_url=local_grand_challenge, verify=False, token=ADMIN_TOKEN)
    # Empty response, but it didn't error out so the server is responding
    assert c.algorithms.page() == []


def test_chunked_uploads(local_grand_challenge):
    file_to_upload = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "testdata", "rnddata"
    )
    # admin
    c_admin = Client(
        token=ADMIN_TOKEN, base_url=local_grand_challenge, verify=False
    )
    existing_chunks_admin = c_admin(path="chunked-uploads/")["count"]
    c_admin.chunked_uploads.upload_file(file_to_upload)
    assert (
        c_admin(path="chunked-uploads/")["count"] == 1 + existing_chunks_admin
    )

    # retina
    c_retina = Client(
        token=RETINA_TOKEN, base_url=local_grand_challenge, verify=False
    )
    existing_chunks_retina = c_retina(path="chunked-uploads/")["count"]
    c_retina.chunked_uploads.upload_file(file_to_upload)
    assert (
        c_retina(path="chunked-uploads/")["count"]
        == 1 + existing_chunks_retina
    )

    c = Client(token="whatever")
    with pytest.raises(HTTPError):
        c.chunked_uploads.upload_file(file_to_upload)


@pytest.mark.parametrize(
    "files",
    (["image10x10x101.mha"], ["image10x10x10.mhd", "image10x10x10.zraw"]),
)
def test_run_external_algorithm(local_grand_challenge, files, monkeypatch):
    c = Client(
        base_url=local_grand_challenge, verify=False, token=ALGORITHMUSER_TOKEN
    )

    existing_us_count = c.raw_image_upload_sessions.list()["count"]
    existing_algorithm_jobs = c.algorithm_jobs.list()["count"]
    existing_images = c.images.list()["count"]

    def return_correct_latest_ready_image(*args, **kwargs):
        algorithm_detail = c.algorithms.page()[0]
        algorithm_detail["latest_ready_image"] = algorithm_detail[
            "algorithm_container_images"
        ][0]
        return {
            "count": 1,
            "next": "",
            "previous": "",
            "results": [algorithm_detail],
        }

    monkeypatch.setattr(
        c.algorithms, "list", return_correct_latest_ready_image
    )
    us = c.run_external_algorithm(
        "Test Algorithm",
        [Path(__file__).parent / "testdata" / f for f in files],
    )

    for _ in range(60):
        us = c.raw_image_upload_sessions.detail(us["pk"])
        if us["status"] == "Succeeded":
            break
        else:
            sleep(0.5)
    else:
        raise TimeoutError

    assert c.raw_image_upload_sessions.list()["count"] == 1 + existing_us_count
    assert c.algorithm_jobs.list()["count"] == 1 + existing_algorithm_jobs
    assert c.images.list()["count"] == 1 + existing_images
